

class Authentication(object):
    def authenticate(self, request, endpoint):
        raise NotImplementedError("Need to implement in a base class")

class DefaultAuthentication(Authentication):
    def authenticate(self, request, endpoint):
        return request.user.is_authenticated()

class NoAuthentication(Authentication):
    def authenticate(self, request, endpoint):
        return True




class Val(object):
    pass


def magic_enum_meta_cls(typ, *args, **kwargs):
    attrs = {}
    for key, value in args[1].items():
        if isinstance(value, Val):
            attrs[key] = key
        else:
            attrs[key] = value
    return type(typ, args[0], attrs, **kwargs)


class MagicEnum(object):
    __metaclass__ = magic_enum_meta_cls




from datetime import datetime

class ApiField(object):
    def __init__(self, name, label=None, obj_attr_name=None, help_text='', default=None):
        self.name = name
        self.obj_attr_name = obj_attr_name if obj_attr_name else name
        self.label = label if label else name
        
    def obj_to_dict(self, obj, data):
        data[self.name] = getattr(obj, self.obj_attr_name)

    def dict_to_obj(self, data, obj):
        if self.name in data:
            setattr(obj, self.obj_attr_name, data[self.name])

    @classmethod
    def type_name(cls):
        return cls.__name__

    @property
    def db_field_name(self):
        return self.name

class CharField(ApiField):
    pass    

class IntegerField(ApiField):
    pass

class TextField(ApiField):
    pass

class DateTimeField(ApiField):
    DATE_FORMAT = '%Y-%m-%d %H:%M:%S'
    def obj_to_dict(self, obj, data):
        val = getattr(obj, self.obj_attr_name)
        # and not isinstance(val, basestring):
        if val != None:
            val = val.strftime(self.DATE_FORMAT)
        data[self.name] = val

    def dict_to_obj(self, data, obj):
        if not self.name in data:
            return
        val = data.get(self.name)
        if val != None and not isinstance(val, datetime):
            val = datetime.strptime(val, self.DATE_FORMAT)
        setattr(obj, self.obj_attr_name, val)


class SimpleForeignKeyField(ApiField):
    def __init__(self, name, label=None, obj_attr_name=None):
        self.name = name + '_id'
        self.obj_attr_name = obj_attr_name if obj_attr_name else self.name
        self.label = label if label else name




from .base_resource import EndPoint, UserError, ArgFilters, GET, PUT, DELETE, POST, BaseApiResource
from .utils import Val, magic_enum_meta_cls
from .fields import DateTimeField, ApiField, SimpleForeignKeyField, CharField, IntegerField, TextField
from django.db import models


django_field_to_sprocket_field = {
    models.DateTimeField.__name__: DateTimeField,
    models.ForeignKey.__name__: SimpleForeignKeyField,
    models.CharField.__name__: CharField,
    models.TextField.__name__: TextField,
    models.IntegerField.__name__: IntegerField,
    models.BigIntegerField.__name__: IntegerField,
    models.PositiveIntegerField.__name__: IntegerField,
}


class DjangoModelResource(BaseApiResource):

    def get_endpoints(self):
        endpoints = []
        endpoints += [
            EndPoint(
                r"^(?P<resource_name>%s)/$" % self._meta.resource_name,
                GET('paged_list', ArgFilters.all_from_query),
                POST('create', ArgFilters.all_from_json),
                ),
            EndPoint(
                r"^(?P<resource_name>%s)/(?P<pk>[0-9a-zA-Z\-_]+)/$" % self._meta.resource_name,
                GET('get', ArgFilters.all_from_query),
                PUT('update', ArgFilters.all_from_json),
                DELETE('delete'),
                ),
        ]
        return endpoints

    def on_init_fields(self, fields):
        for field in self._meta.model_class._meta.fields:
            cls = django_field_to_sprocket_field.get(field.__class__.__name__, ApiField)
            api_field = cls(field.name)
            fields.append(api_field)

    def create(self, **kwargs):
        obj = self.dict_to_obj(kwargs)
        return self.create_object(obj)

    def create_object(self, obj):
        self.execute_handlers(ModelEvents.pre_save_prepare, obj)
        self.execute_handlers(ModelEvents.pre_create_prepare, obj)

        self.execute_handlers(ModelEvents.pre_save_validate, obj)
        self.execute_handlers(ModelEvents.pre_create_validate, obj)

        self.execute_handlers(ModelEvents.pre_process_save, obj)
        self.execute_handlers(ModelEvents.pre_process_update, obj)

        obj.save()

        self.execute_handlers(ModelEvents.post_save, obj)
        self.execute_handlers(ModelEvents.post_create, obj)

        return obj

    def update_object(self, obj):
        data = self.obj_to_dict(obj)
        return self.update(obj.pk, **data)

    def update(self, pk, **kwargs):
        obj = self.get(pk=pk)
        if obj == None:
            return self.handle_update_not_found(pk, **kwargs)
        previous_data = {}
        for name in kwargs.keys():
            previous_data[name] = getattr(obj, name, None)
        obj = self.dict_to_obj(kwargs, obj)

        self.execute_handlers(ModelEvents.pre_save_prepare, obj)
        self.execute_handlers(ModelEvents.pre_update_prepare, obj, previous_data)

        self.execute_handlers(ModelEvents.pre_save_validate, obj)
        self.execute_handlers(ModelEvents.pre_update_validate, obj, previous_data)

        self.execute_handlers(ModelEvents.pre_process_save, obj)
        self.execute_handlers(ModelEvents.pre_process_update, obj, previous_data)

        obj.save()

        self.execute_handlers(ModelEvents.post_save, obj)
        self.execute_handlers(ModelEvents.post_update, obj, previous_data)

        return obj

    def handle_update_not_found(self, pk, **kwargs):
        raise UserError("Object not found for key %s" % pk, 404)

    def delete(self, pk):
        obj = self.get(pk=pk)
        if not obj:
            raise UserError("Object with key %s was not found, could not delete." % pk, 404)
        self.execute_handlers(ModelEvents.delete_prepare, obj)
        self.execute_handlers(ModelEvents.delete_validate, obj)
        self.execute_handlers(ModelEvents.delete_process, obj)
        obj.delete()
        self.execute_handlers(ModelEvents.post_delete, obj)
        return True

    def get(self, **kwargs):
        items = self._list_and_count(limit=1, _include_total=False, **kwargs)
        if items:
            self.execute_handlers(ModelEvents.get_object, items[0])
            return items[0]
        else:
            return None

    def paged_list(self, offset=0, limit=20, **kwargs):
        items, total = self._list_and_count(offset=int(offset), limit=int(limit), **kwargs)
        data = {
            'offset': int(offset),
            'limit': int(limit),
            'total_count': total,
            'objects': self.obj_list_to_dicts(items)
            }
        return data

    def list(self, **kwargs):
        return self._list_and_count(_include_total=False, **kwargs)

    def _list_and_count(self, offset=0, limit=None, _include_total=True, **kwargs):
        q_filters, filters = build_django_orm_filters_from_params(self, kwargs)
        self.execute_handlers(ModelEvents.adjust_orm_filters, q_filters, filters)
        queryset = self._meta.model_class.objects.filter(*q_filters, **filters)
        queryset = self.execute_filters(ModelEvents.chain_queryset, queryset)
        if limit != None:
            items = queryset[offset:offset + limit]
        else:
            items = list(queryset)
        items = self.execute_filters(ModelEvents.filter_objects, items)
        items = list(items)
        self.execute_handlers(ModelEvents.list_objects, items)
        if _include_total:
            return items, queryset.count()
        return items


def build_django_orm_filters_from_params(api_resource, params):
    '''
    Takes filters that came in from a query string and turns them into
    django ORM filters
    '''
    q_filters = []
    dj_filters = {}
    for key, value in params.items():
        parts = key.split('__')
        field_name = parts.pop(0)
        if field_name not in api_resource.field_by_name and not field_name == 'pk':
            continue
        filter_type = 'exact'
        if parts:
            filter_type = parts[-1]
        validate_filter(api_resource, field_name, filter_type)

        if value in [True, 'true', 'True']:
            value = True
        elif value in [False, 'false', 'False']:
            value = False
        elif value in (None, 'nil', 'none', 'None'):
            value = None

        if filter_type == 'ne':
            q_filters.append(
                ~models.Q(**{field_name: value})
            )
            continue

        filter_key = field_name + '__' + filter_type

        is_container = type(value) in (list, set, frozenset, tuple)
        # if we're doing an "in" query, all queries need to be in an iterable container, even if one variable.
        if filter_type in ["in", "range"]:
            if not is_container:
                value = [value]
        else:
            if is_container:
                if len(value) > 0:
                    value = value[0]
                else:
                    value = None

        # Hack to fix filtering of foreign keys, which need to be filtered
        # on the field without the _id part addded
        if isinstance(api_resource.field_by_name.get(field_name), SimpleForeignKeyField):
            if field_name.endswith('_id') and filter_type == 'exact':
                filter_key = field_name[:-3] + '__exact'

        dj_filters[filter_key] = value

    return q_filters, dj_filters


def validate_filter(api, field_name, filter_type):
    if not field_name in api._meta.filtering:
        raise UserError('You cannot filter on the field %s ' % field_name, status_code=400)
    if not filter_type in api._meta.filtering.get(field_name, []):
        raise UserError('You cannot filter on the field %s in the form of %s ' % (field_name, filter_type), 400)


class ModelEvents(object):
    __metaclass__ = magic_enum_meta_cls

    pre_save_prepare = Val()
    pre_save_validate = Val()
    pre_process_save = Val()
    post_save = Val()

    pre_create_prepare = Val()
    pre_create_validate = Val()
    pre_process_create = Val()
    post_create = Val()

    pre_update_prepare = Val()
    pre_update_validate = Val()
    pre_process_update = Val()
    post_update = Val()

    adjust_orm_filters = Val()
    chain_queryset = Val()
    filter_objects = Val()
    list_objects = Val()
    get_object = Val()

    delete_prepare = Val()
    delete_validate = Val()
    delete_process = Val()
    post_delete = Val()



from datetime import datetime
import inspect
import time
from unittest import TestCase

from django.conf import settings
from django.conf.urls.defaults import patterns, include, url
from django.test.client import Client
from django.utils import simplejson

from mocking_bird.mocking import MockingBirdMixin

from ..mixins import BaseMixin
from ..auth import NoAuthentication
from ..fields import DateTimeField, ApiField
from ..base_resource import BaseApiResource, ResourceMeta, EndPoint, ArgFilters, POST, PUT, GET, UserError, UnauthenticatedError


class SimpleCase(TestCase, MockingBirdMixin):
    def test_direct_crud(self):
        resource = SimpleResource()
        dt = datetime(2011, 4, 1, 12, 30, 0)
        label = 'ATestLabel'
        nicknames = ['jack', 'jacko']
        obj = resource.create(label=label, published_at=dt, nicknames=nicknames)

        obj2 = resource.get(pk=obj.pk)
        self.assertEquals(label, obj2.label)
        self.assertEquals(nicknames, obj2.nicknames)

        new_label = 'ANewLabel'
        resource.update(obj.pk, label=new_label)
        obj3 = resource.get(pk=obj.pk)
        self.assertEquals(new_label, obj3.label)

    def test_http_crud(self):
        c = Client()

        dt = datetime(2011, 4, 1, 12, 30, 0)
        dt_str = dt.strftime('%Y-%m-%d %H:%M:%S')
        label = 'ATestLabel'
        nicknames = ['jack', 'jacko']

        post_data = {
            'label': label,
            'published_at': dt_str,
            'nicknames': nicknames
            }

        r = c.post(
            '/api/simple-resource',
            data=simplejson.dumps(post_data),
            content_type='application/json')
        self.assertEquals(200, r.status_code)
        data = simplejson.loads(r.content)
        self.assertTrue(data['pk'] > 10000)
        self.assertEquals(r['x-sprocket-new-object-id'], str(data['pk']))

        r = c.get(
            '/api/simple-resource/%s' % data['pk'],
            content_type='application/json')
        self.assertEquals(200, r.status_code)
        data = simplejson.loads(r.content)
        self.assertEquals(label, data['label'])

        new_label = 'NewLabel'
        post_data['label'] = new_label
        r = c.put(
            '/api/simple-resource/%s' % data['pk'],
            data=simplejson.dumps(post_data),
            content_type='application/json')
        data = simplejson.loads(r.content)
        self.assertEquals(200, r.status_code)
        self.assertEquals(new_label, data['label'])

        r = c.get(
            '/api/simple-resource',
            content_type='application/json')
        self.assertEquals(200, r.status_code)
        data = simplejson.loads(r.content)
        self.assertEquals(1, len(data))

    def test_mixins(self):
        c = Client()

        label = 'ATestLabel'

        post_data = {
            'label': label,
            }

        r = c.post(
            '/api/simple-resource',
            data=simplejson.dumps(post_data),
            content_type='application/json')
        data = simplejson.loads(r.content)

        r = c.post(
            '/api/simple-resource/%s/soft-delete' % data['pk'],
            content_type='application/json')
        self.assertEquals(200, r.status_code)

        r = c.get(
            '/api/simple-resource/%s' % data['pk'],
            content_type='application/json')
        self.assertEquals(200, r.status_code)
        data = simplejson.loads(r.content)
        # The created timestamp should not have the default 1900 value
        self.assertTrue('1900' not in data['created'] and '201' in data['created'])
        # Verify the soft-delete worked
        self.assertTrue(data['deleted'] == True)

    def test_error_handling(self):
        c = Client()

        label = 'ATestLabel'

        post_data = {
            'label': label,
            }
        r = c.post(
            '/api/simple-resource',
            data=simplejson.dumps(post_data),
            content_type='application/json')
        data = obj_data = simplejson.loads(r.content)
        self.assertEquals(200, r.status_code)

        r = c.put(
            '/api/simple-resource/%s' % data['pk'],
            data=simplejson.dumps({'nicknames': ['a', 'b', 'c', 'd', 'e', 'f']}),
            content_type='application/json')
        data = simplejson.loads(r.content)
        self.assertEquals(400, r.status_code)

        # Illegal method
        r = c.delete(
            '/api/simple-resource',
            content_type='application/json')
        self.assertEquals(405, r.status_code)

        r = c.get('/api/simple-resource/%s?denyMe=true' % obj_data['pk'])
        self.assertEquals(403, r.status_code)

    url_conf = 'sprocket.test.test_base_resource'

    def setUp(self):
        super(SimpleCase, self).setUp()
        self.org_urls = settings.ROOT_URLCONF
        settings.ROOT_URLCONF = self.url_conf
        simple_resource._storage = {}

    def tearDown(self):
        super(SimpleCase, self).tearDown()
        settings.ROOT_URLCONF = self.org_urls


class SimpleObject(object):
    pk = 0
    label = ''
    city = ''
    created = datetime(1900, 1, 1)
    updated = datetime(1900, 1, 1)
    published_at = datetime.utcnow()
    nicknames = ()
    deleted = False

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class SimpleResource(BaseApiResource):
    _storage = {}

    class Meta(ResourceMeta):
        resource_name = 'simple-resource'
        model_class = SimpleObject

    def on_init_fields(self, fields):
        for name in dir(SimpleObject):
            attr = getattr(SimpleObject, name)
            if name.startswith('_'):
                continue
            if callable(attr):
                continue
            if isinstance(attr, datetime):
                cls = DateTimeField
            else:
                cls = ApiField
            fields.append(cls(name))

    def get_endpoints(self):
        endpoints = [
            EndPoint(
                r"^(?P<resource_name>%s)$" % self._meta.resource_name,
                GET('list', ArgFilters.fields_from_query),
                POST('create', ArgFilters.fields_from_json)
                ),
            EndPoint(
                r"^(?P<resource_name>%s)/(?P<pk>[\d]+)$" % self._meta.resource_name,
                GET('get', ArgFilters.fields_from_query),
                PUT('update', ArgFilters.fields_from_json)
                ),
            ]
        return endpoints

    def get_mixins(self):
        return [DeletedUpdatedMixin(self)]

    def update(self, pk, **kwargs):
        if len(kwargs.get('nicknames', [])) > 5:
            raise UserError("Too many nicknames", 400)
        obj = self._storage[long(pk)]
        for field in self.fields:
            if field.name in kwargs:
                field.dict_to_obj(kwargs, obj)
        self.execute_handlers('updated', obj)
        return obj

    def get(self, pk):
        o = self._storage[long(pk)]
        return o

    def list(self, **kwargs):
        items = []
        for item in self._storage.values():
            filter_out = False
            if kwargs:
                for key, val in kwargs.items():
                    if getattr(item, 'key', None) != val:
                        filter_out = True
                        break
            if not filter_out:
                items.append(item)
        return items

    def create(self, **kwargs):
        obj = SimpleObject()
        for field in self.fields:
            if field.name in kwargs:
                field.dict_to_obj(kwargs, obj)
        obj.pk = long(time.time() * 1000)
        self.execute_handlers('created', obj)
        self._storage[obj.pk] = obj
        self.set_response_header('x-sprocket-new-object-id', str(obj.pk))
        return obj

    def delete(self, pk):
        del self._storage[long(pk)]
        self.execute_handlers('deleted', long(pk))

    def on_authenticate(self, request):
        if request.GET.get('denyMe') == 'true':
            raise UnauthenticatedError("denyMe was true")


class DeletedUpdatedMixin(BaseMixin):
    def get_endpoints(self):
        return [
            EndPoint(
                r"^(?P<resource_name>%s)/(?P<pk>[\d]+)/soft-delete$" % self.api._meta.resource_name,
                POST('soft_delete'))
            ]

    def on_created(self, obj):
        obj.created = datetime.utcnow()

    def soft_delete(self, pk):
        self.api._storage[long(pk)].deleted = True
        return True


simple_resource = SimpleResource()
urlpatterns = patterns('',
    (r'^api/', include(simple_resource.urls)),
)








class BaseMixin(object):
    def __init__(self, api_resource):
        self.api = api_resource
        self.event_handler_by_event_name = self._construct_events_dict()
        
    def _construct_events_dict(self):
        handlers_by_name = {}
        for attr_name in dir(self):
            if attr_name.startswith('on_'):
                event_name = attr_name[3:]
                handlers_by_name[event_name] = getattr(self, attr_name)
        return handlers_by_name

    def get_handler_for_event(self, event_name):
        return self.event_handler_by_event_name.get(event_name, None)

    def get_endpoints(self):
        return []




DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3', # Add 'postgresql_psycopg2', 'postgresql', 'mysql', 'sqlite3' or 'oracle'.
        'NAME': 'sprocket.test.db',                      # Or path to database file if using sqlite3.
        'USER': '',                      # Not used with sqlite3.
        'PASSWORD': '',                  # Not used with sqlite3.
        'HOST': '',                      # Set to empty string for localhost. Not used with sqlite3.
        'PORT': '',                      # Set to empty string for default. Not used with sqlite3.
    }
}



DEBUG = True
TEMPLATE_DEBUG = True

TIME_ZONE = 'America/New_York'

INSTALLED_APPS = (
    'sprocket',
)

CACHES = {
    'default': {
        'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
        'LOCATION': 'cosdev'
        }
    }    

ROOT_URLCONF = ''




import inspect
import threading
import traceback

from django.conf.urls.defaults import *
from django.http import HttpResponse, HttpRequest, HttpResponseNotAllowed
from django.utils import simplejson as json
from django.views.decorators.csrf import csrf_exempt

from .utils import MagicEnum, Val, magic_enum_meta_cls
from .auth import DefaultAuthentication
from .fields import ApiField


class ResourceMeta(object):
    resource_name = None
    model_class = object
    includes = []
    excludes = []
    filtering = None

    def __init__(self):
        if not self.filtering:
            self.filtering = {}
        if not 'id' in self.filtering:
            self.filtering['id'] = ['exact']
        if not 'pk' in self.filtering:
            self.filtering['pk'] = ['exact']


class BaseApiResource(object):
    _meta = ResourceMeta()  # Overwritten by the __new__ method, but kept here so pylint can do type inference

    class Meta(ResourceMeta):
        resource_name = 'base-api-resource'
        authentication = DefaultAuthentication()

    def __new__(typ, *args, **kwargs):
        obj = object.__new__(typ, *args, **kwargs)
        obj._meta = obj.Meta()
        return obj

    def __init__(self, *args, **kwargs):
        self._event_handlers = {}
        self._merge_in_mixins()
        self.fields = self._init_fields()
        self.field_names = [field.name for field in self.fields]
        self.field_by_name = dict([(field.name, field) for field in self.fields])
        self.urls = self._build_urls()

    # Initialize endpoints and mixins
    def _merge_in_mixins(self):
        '''
        All the methods in mixins that do not being with 'on_' or '_' get mixed into the API
        class - however we have to do some trickery so that when they are called 'self' is the
        'self' of the mixin, not of the API Resource.
        '''
        self.mixins = self.get_mixins()
        self.mixins_by_name = {}
        for mixin in self.mixins:
            mixin.api = self
            self.mixins_by_name[mixin.__class__.__name__] = mixin
            for attr_name in dir(mixin):
                if not isinstance(attr_name, basestring):
                    continue
                if attr_name.startswith('on_'):
                    continue
                if attr_name == 'get_endpoints':
                    continue
                if attr_name.startswith('_'):
                    continue
                attr = getattr(mixin, attr_name)
                if inspect.ismethod(attr):
                    wrapped_method = self._wrap_mixin_method(attr)
                    setattr(self, attr_name, wrapped_method)

    def _wrap_mixin_method(self, attr, *args, **kwargs):
        def func(*args, **kwargs):
            return attr(*args, **kwargs)
        return func

    def get_mixins(self):
        # Override in subclasses
        return []

    def _build_urls(self):
        urls = []
        for endpoint in self._get_all_endpoints():
            urls.append(url(
                endpoint.url_pattern,
                self.wrap(endpoint),
                name=endpoint.name))
        return urls

    def _get_all_endpoints(self):
        endpoints = []
        endpoints.extend(self.get_override_endpoints())
        for mixin in self.mixins:
            endpoints.extend(mixin.get_endpoints())
        endpoints.extend(self.get_endpoints())

        return endpoints

    def get_endpoints(self):
        # Override me in a subclass, return a list of Endpoint objects
        raise NotImplementedError()

    def get_override_endpoints(self):
        # Override in a subclass - these endpoints have precedence over any mixin endpoints
        return []

    # Dispatching a request

    def wrap(self, endpoint):
        @csrf_exempt
        def handler(request, **kwargs):
            try:
                _thread_local.current = CurrentRequestThreadHolder(request)
                request.endpoint = endpoint
                return self._dispatch(endpoint, request, kwargs)
            except UserError, ex:
                return self.handle_user_error(request, endpoint, ex, ex.to_response())
            except ApiError, ex:
                return self.handle_server_error(request, endpoint, ex, ex.to_response())
            except Exception, ex:
                return self.handle_server_error(
                    request,
                    endpoint,
                    ex,
                    HttpResponse(
                        json.dumps({'message': 'There was an internal error'}),
                        status=500
                    ))
            finally:
                _thread_local.current = None
        return handler

    def _dispatch(self, endpoint, request, kwargs):
        self._authenticate(request)
        if not request.method in endpoint.http_method_dict:
            return HttpResponseNotAllowed(endpoint.http_method_dict.keys())
        method_endpoint = request.method_endpoint = endpoint.http_method_dict[request.method]
        if isinstance(method_endpoint.api_method_name, basestring):
            method = getattr(self, method_endpoint.api_method_name)
        else:
            method = method_endpoint.api_method_name
        self._adjust_kwargs(method_endpoint, request, kwargs)
        self.execute_handlers(BaseEvents.pre_dispatch_request, request, kwargs)
        result = method(**kwargs)
        response = self._result_to_response(result)
        self.add_preset_response_info(response)
        self.execute_handlers(BaseEvents.process_response, response)
        return response

    def _authenticate(self, request):
        self.execute_handlers(BaseEvents.authenticate, request)

    def _adjust_kwargs(self, method_endpoint, request, kwargs):
        if 'resource_name' in kwargs:
            del kwargs['resource_name']
        if 'api_name' in kwargs:
            del kwargs['api_name']

        self.execute_handlers(BaseEvents.adjust_kwargs_for_request, request, kwargs)

        for filter in method_endpoint.arg_filters:
            filter(self, request, kwargs)

    def _result_to_response(self, result):
        response = self.current_request.method_endpoint.to_response_func(result)
        if response:
            return response
        if isinstance(result, HttpResponse):
            return result
        status_code = 200
        if self.current_response_status_code != None:
            status_code = self.current_response_status_code
        if isinstance(result, basestring):
            return HttpResponse(result, status=status_code)
        elif isinstance(result, dict):
            return HttpResponse(json.dumps(result), status=status_code)
        elif isinstance(result, list) and len(result) > 0 and isinstance(result[0], self._meta.model_class):
            return HttpResponse(self.obj_list_to_str(result), status=status_code)
        elif isinstance(result, list):
            return HttpResponse(json.dumps(result), status=status_code)
        elif isinstance(result, self._meta.model_class):
            return HttpResponse(self.obj_to_str(result), status=status_code)
        elif result == None and self.current_request.method == 'GET':
            return HttpResponse('{"message": "Object not found"}', status=404)
        elif result == True:
            return HttpResponse('{"succeeded": true, "message": "Action succeeded"}', status=status_code)
        else:
            raise Exception("Response was of an unexpected type")

    def add_preset_response_info(self, response):
        '''
        Adds any headers or cookies that were set during the processing of the api call
        '''
        for key, val in _thread_local.current.response_headers.items():
            response[key] = val
        for args, kwargs in _thread_local.current.cookie_setters:
            response.set_cookie(*args, **kwargs)
        if 'content-type' not in _thread_local.current.response_headers:
            response['Content-type'] = 'application/json'

    def handle_user_error(self, request, endpoint, exc, response):
        ''' Override this to add any custom error reporting logic '''
        return response

    def handle_server_error(self, request, endpoint, exc, response):
        ''' Override this to add any custom error reporting logic '''
        traceback.print_exc()
        return response

    def execute_handlers(self, event_name, *args):
        for handler in self._get_handlers(event_name):
            handler(*args)

    def execute_filters(self, event_name, items, *args):
        for handler in self._get_handlers(event_name):
            items = handler(items, *args)
        return items

    def _get_handlers(self, event_name):
        if event_name not in self._event_handlers:
            self._event_handlers[event_name] = self._build_handlers_list(event_name)
        return self._event_handlers[event_name]

    def _build_handlers_list(self, event_name):
        handlers = []
        for mixin in self.mixins:
            handler = mixin.get_handler_for_event(event_name)
            if handler:
                handlers.append(handler)
        handler = getattr(self, 'on_' + event_name, None)
        if handler:
            handlers.append(handler)
        return handlers

    # Fields and Serialization
    def _init_fields(self):
        fields = []
        self.execute_handlers(BaseEvents.init_fields, fields)
        return fields

    def obj_to_str(self, obj):
        data = self.obj_to_dict(obj)
        return self.serialize(data)

    def serialize(self, data):
        return json.dumps(data)

    def obj_to_dict(self, obj):
        includes = self._meta.includes
        excludes = self._meta.excludes
        data = {}
        for field in self.fields:
            if field.name in excludes:
                continue
            if includes and field.name not in includes:
                continue
            field.obj_to_dict(obj, data)
        self.execute_handlers(BaseEvents.obj_to_dict, obj, data)
        return data

    def str_to_obj(self, post_data_byte_str):
        data = self.deserialize(post_data_byte_str)
        return self.dict_to_obj(data)

    def deserialize(self, data_str):
        try:
            return json.loads(data_str)
        except ValueError, e:
            raise UserError(e.message, status_code=400)

    def dict_to_obj(self, data, obj=None):
        includes = self._meta.includes
        excludes = self._meta.excludes
        if obj == None:
            obj = self._meta.model_class()
        for field in self.fields:
            if field.name in excludes:
                continue
            if includes and field.name not in includes:
                continue
            field.dict_to_obj(data, obj)
        self.execute_handlers(BaseEvents.dict_to_obj, data, obj)
        return obj

    def obj_list_to_str(self, objects):
        return json.dumps(self.obj_list_to_dicts(objects))

    def obj_list_to_dicts(self, objects):
        return [self.obj_to_dict(obj) for obj in objects]

    # Helpers - we store the request object in thread local so we don't pass it around everywhere

    @property
    def current_request(self):
        req = self.thread_current.request
        if not req:
            req = EmptyRequest()
        return req

    @property
    def current_user(self):
        return self.current_request.user

    @property
    def current_username(self):
        user = self.current_user
        if user == None or not user.username:
            return ''
        else:
            return user.username

    @property
    def current_response_status_code(self):
        return self.thread_current.status_code

    def set_current_status_code(self, status_code):
        self.thread_current.status_code = status_code

    def set_response_header(self, name, value):
        self.thread_current.response_headers[name.lower()] = value

    def set_response_cookie(self, *args, **kwargs):
        self.thread_current.cookie_setters.append((args, kwargs))

    @property
    def thread_current(self):
        c = getattr(_thread_local, 'current', None)
        if c == None:
            c = CurrentRequestThreadHolder(None)
        return c

    # Default event handlers, to be overridden in subclasses
    def on_authenticate(self, request):
        if not request.user.is_authenticated():
            raise UnauthenticatedError("This is not an authenticated request")


class EmptyRequest(HttpRequest):
    def __nonzero__(self):
        return False


class ArgFilters(object):
    '''
    Filters that take the incoming request data and generate
    the appropriate keyword arguments that will be passed to the api method
    '''

    @staticmethod
    def all_from_json(api, request, kwargs):
        return ArgFilters.all_from_json_custom()(api, request, kwargs)

    @staticmethod
    def all_from_json_custom(allow_empty=False):
        def filter_func(api, request, kwargs):
            if request.method in ('POST', 'PUT'):
                data = ArgFilters.get_json_data(request, allow_empty)
                if data:
                    kwargs.update(data)
        return filter_func

    @staticmethod
    def fields_from_json(api, request, kwargs):
        if request.method in ('POST', 'PUT'):
            data = ArgFilters.get_json_data(request)
            for field in api.fields:
                if field.name in data:
                    kwargs[field.name] = data[field.name]

    @staticmethod
    def from_keys(keys):
        if not hasattr(keys, '__iter__'):
            keys = [keys]

        def filter_func(api, request, kwargs):
            data = ArgFilters.get_json_data(request)
            for key in keys:
                if key not in data:
                    raise UserError("Expected key %s in the posted json data" % key, 400)
                kwargs[key] = data[key]

        return filter_func

    @staticmethod
    def keys_from_query(keys, all_required=True):
        def filter_func(api, request, kwargs):
            for key in keys:
                if key not in request.GET:
                    if all_required:
                        raise UserError("Expected key %s in the posted json data" % key)
                else:
                    values = request.GET.getlist(key)
                    if len(values) > 1:
                        kwargs[key] = values
                    elif len(values) == 1:
                        kwargs[key] = values[0]
        return filter_func

    @staticmethod
    def fields_from_query_plus(additional_keys):
        def filter_func(api, request, kwargs):
            return ArgFilters.fields_from_query(api, request, kwargs, additional_keys=additional_keys)
        return filter_func

    @staticmethod
    def fields_from_query(api, request, kwargs, additional_keys=('offset', 'limit', 'order_by')):
        field_names = set([field.name for field in api.fields])
        for name, values in request.GET.lists():
            if name in field_names or name in additional_keys:
                if len(values) > 1:
                    kwargs[name] = values
                elif len(values) == 1:
                    kwargs[name] = values[0]
                continue
            # We also allow queries in the form 'field_name__gt', 'field_name__lt' that get passed into filters
            i = name.find('__')
            if i > -1:
                if name[:i] in field_names:
                    if len(values) > 1:
                        kwargs[name] = values
                    elif len(values) == 1:
                        kwargs[name] = values[0]

    _internal_query_keys = ['portalId', 'hapikey', 'scopes']

    @staticmethod
    def all_from_query(api, request, kwargs):
        for name, values in request.GET.lists():
            if name not in ArgFilters._internal_query_keys:
                if len(values) > 1:
                    kwargs[name] = values
                elif len(values) == 1:
                    kwargs[name] = values[0]

    @staticmethod
    def with_data(api, request, kwargs):
        try:
            data = json.loads(request.raw_post_data)
        except Exception:
            raise UserError("Invalid syntax for the json data", status_code=400)
        kwargs['data'] = data

    @staticmethod
    def get_json_data(request, allow_empty=False):
        post_data = request.raw_post_data
        if allow_empty and \
           (post_data is None or len(post_data) == 0 or post_data == ''):
            return None
        try:
            data = json.loads(post_data)
        except Exception:
            raise UserError("Invalid syntax for the json data", status_code=400)
        return data

    @staticmethod
    def convert_to_bool_values(param_names=None):
        '''
        Returns a filter function to parse parameters (all or just the given ones) as specific types
        @param_names - (optional) - string names of boolean-valued params
        '''
        def filter_func(api, request, kwargs):
            for key, val in kwargs.items():
                if param_names == None or (param_names and key in param_names):
                    if val == 'true':
                        kwargs[key] = True
                    elif val == 'false':
                        kwargs[key] = False

        return filter_func

    @staticmethod
    def convert_to_int_values(param_names=None):
        '''
        Returns a filter function to parse parameters (all or just the given ones) as integer values
        @param_names - (optional) - string names of integer-valued params
        '''
        def filter_func(api, request, kwargs):
            for key, val in kwargs.items():
                # Ensure object is a string before trying to determine if its content is of integer value
                if val and hasattr(val, 'isdigit') and val.isdigit() and \
                   (param_names == None or (param_names and key in param_names)):
                    try:
                        kwargs[key] = int(val)
                    except:
                        pass
        return filter_func

    @staticmethod
    def convert(**keys_to_callable):
        '''
        k1=fn1,
        k2=fn2,
        ...
        '''
        def filter_func(api, request, kwargs):
            for key, val in kwargs.items():
                # Ensure object is a string before trying to determine if its content is of integer value
                if key not in keys_to_callable:
                    continue

                try:
                    kwargs[key] = keys_to_callable[key](val)
                except:
                    pass

        return filter_func


class EndPoint(object):
    def __init__(
        self,
        url_pattern,
        *end_point_methods,
        **kwargs):
        self.url_pattern = url_pattern
        self.http_method_dict = {}
        for epm in end_point_methods:
            self.http_method_dict[epm.__class__.__name__.upper()] = epm
        self.name = kwargs.get('name', '')


class EndPointMethod(object):
    '''
    Defines the endpoint handler for a particular HTTP Method
    '''
    def __init__(self, api_method_name, arg_filters=(), to_response_func=None):
        '''
        @api_method_name - a name a method that exists on the API resource that will be the handler for this endpoint
        @arg_filters - a list of filter functions that will pull keys and values of the request and include them as kwargs
        @to_response_func - optional if included will turn the result of the api_method into an HttpResponse object
        '''
        if callable(arg_filters):
            arg_filters = [arg_filters]
        if to_response_func == None:
            to_response_func = lambda o: None

        self.api_method_name = api_method_name
        self.to_response_func = to_response_func
        self.arg_filters = arg_filters


class PUT(EndPointMethod):
    pass


class POST(EndPointMethod):
    pass


class GET(EndPointMethod):
    pass


class DELETE(EndPointMethod):
    pass


class BaseEvents(object):
    __metaclass__ = magic_enum_meta_cls

    authenticate = Val()
    adjust_kwargs_for_request = Val()
    pre_dispatch_request = Val()
    process_response = Val()
    dict_to_obj = Val()
    obj_to_dict = Val()
    init_fields = Val()


class ApiError(Exception):
    def __init__(self, message, status_code, error_dict=None, errors=(), extra_data=None):
        super(ApiError, self).__init__(message)
        self.message = message
        self.status_code = status_code
        self.error_dict = error_dict
        self.extra_data = extra_data
        self.errors = errors

    def get_data(self):
        d = {'message': self.message, 'succeeded': False}
        if self.error_dict != None:
            d['error_dict'] = self.error_dict
        if self.errors != None:
            d['errors'] = self.errors

        if self.extra_data != None:
            d.update(self.extra_data)
        return d

    def to_response(self):
        return HttpResponse(json.dumps(self.get_data()), status=self.status_code)


class UserError(ApiError):
    pass


class UnauthenticatedError(UserError):
    def __init__(self, message, status_code=403):
        super(UnauthenticatedError, self).__init__(message, status_code)


class CurrentRequestThreadHolder(object):
    def __init__(self, request):
        self.request = request
        self.response_headers = {}
        self.cookie_setters = []
        self.status_code = None

_thread_local = threading.local()
_thread_local.current = CurrentRequestThreadHolder(None)



from setuptools import setup, find_packages

setup(
    name='sprocket',
    version='0.1',
    description='Web API organization with Mixins.',
    author='Patrick Fitzsimmons',
    author_email='pfitzsimmons@hubspot.com',
    url=' http://hubspot.com/',
    packages=find_packages(),
    install_requires=[
        'Django',
        ],
    platforms=["any"],
)



from datetime import datetime
import time
from unittest import TestCase

from django.conf import settings
from django.conf.urls.defaults import patterns, include
from django.db.models import Model, CharField, DateTimeField as DjDateTimeField, EmailField, IntegerField
from django.db import connections, transaction
from django.test.client import Client
from django.utils import simplejson

from mocking_bird.mocking import MockingBirdMixin

from ..django_model_resource import DjangoModelResource
from ..base_resource import ResourceMeta


class SimpleCase(TestCase, MockingBirdMixin):
    def test_direct_crud(self):
        label = 'MyLabelz'
        email = 'amail@maila.com'
        second_label = 'MySecondLabel'

        obj = my_resource.create(
            label=label,
            email=email,
            age=17)
        obj = my_resource.get(pk=obj.pk)
        self.assertEquals(label, obj.label)

        obj = my_resource.get(email=email)
        self.assertEquals(label, obj.label)

        second_label = 'MySecondLabel'
        obj = my_resource.create(
            label=second_label,
            age=21)

        objs = my_resource.list()
        self.assertEquals(2, len(objs))

        # test multiple __in
        objs = my_resource.list(age__in=[17, 21])
        self.assertEquals(2, len(objs))

        # test multiple __in with alternative container
        objs = my_resource.list(age__in=set([17, 21]))
        self.assertEquals(2, len(objs))

        # test single __in
        objs = my_resource.list(age__in=[17])
        self.assertEquals(1, len(objs))

        # test single equals
        objs = my_resource.list(age=17)
        self.assertEquals(1, len(objs))

        # test range
        objs = my_resource.list(age__range=[17, 21])
        self.assertEquals(2, len(objs))

        # test gt
        objs = my_resource.list(age__gt=17)
        self.assertEquals(1, len(objs))

        # test gte
        objs = my_resource.list(age__gte=17)
        self.assertEquals(2, len(objs))

        # test lt
        objs = my_resource.list(age__lt=21)
        self.assertEquals(1, len(objs))

        # test lte
        objs = my_resource.list(age__lte=21)
        self.assertEquals(2, len(objs))

    def test_http_crud(self):
        c = Client()

        new_label = 'NewMylabelszz'
        dt = datetime(2011, 7, 1, 12, 30, 0)
        dt_str = dt.strftime('%Y-%m-%d %H:%M:%S')

        post_data = {
            'label': 'MyLabelz',
            'published_at': dt_str,
            'age': 17,
            'email': "amail@maila.com"
            }

        second_post_data = {
            'label': 'MySecondLabel',
            'published_at': dt_str,
            'age': 21,
            'email': 'amail2@maila.com',
        }

        # Create an object
        r = c.post(
            '/api/my-resource/',
            data=simplejson.dumps(post_data),
            content_type='application/json')
        self.assertEquals(200, r.status_code)
        data = simplejson.loads(r.content)
        obj_data = data
        self.assertTrue(data['id'] > 10000)

        # Now get the object we just created
        r = c.get('/api/my-resource/%s/' % data['id'])
        self.assertEquals(200, r.status_code)
        data = simplejson.loads(r.content)
        self.assertTrue(post_data['label'], data['label'])
        self.assertTrue(dt_str, data['published_at'])

        # Update the object
        post_data['label'] = new_label
        r = c.put(
            '/api/my-resource/%s/' % data['id'],
            data=simplejson.dumps(post_data),
            content_type='application/json')
        self.assertEquals(200, r.status_code)
        data = simplejson.loads(r.content)
        self.assertTrue(new_label, data['label'])

        # Post a second object
        r = c.post(
            '/api/my-resource/',
            data=simplejson.dumps(second_post_data),
            content_type='application/json')
        self.assertEquals(200, r.status_code)

        # The list should now have two objects
        r = c.get('/api/my-resource/')
        data = simplejson.loads(r.content)
        self.assertEquals(200, r.status_code)
        self.assertEquals(2, len(data['objects']))
        self.assertEquals(2, data['total_count'])

        r = c.get('/api/my-resource/?offset=1')
        data = simplejson.loads(r.content)
        self.assertEquals(200, r.status_code)
        self.assertEquals(1, len(data['objects']))
        self.assertEquals(2, data['total_count'])

        # check on filtering
        r = c.get('/api/my-resource/?age__in=17&age__in=21')
        data = simplejson.loads(r.content)
        self.assertEquals(200, r.status_code)
        self.assertEquals(2, len(data['objects']))
        self.assertEquals(2, data['total_count'])

        # check on filtering (single value)
        r = c.get('/api/my-resource/?age__in=17')
        data = simplejson.loads(r.content)
        self.assertEquals(200, r.status_code)
        self.assertEquals(1, len(data['objects']))
        self.assertEquals(1, data['total_count'])

        r = c.get('/api/my-resource/?age__gt=17')
        data = simplejson.loads(r.content)
        self.assertEquals(200, r.status_code)
        self.assertEquals(1, len(data['objects']))
        self.assertEquals(1, data['total_count'])

        r = c.get('/api/my-resource/?age__gte=17')
        data = simplejson.loads(r.content)
        self.assertEquals(200, r.status_code)
        self.assertEquals(2, len(data['objects']))
        self.assertEquals(2, data['total_count'])

        r = c.get('/api/my-resource/?age__lt=21')
        data = simplejson.loads(r.content)
        self.assertEquals(200, r.status_code)
        self.assertEquals(1, len(data['objects']))
        self.assertEquals(1, data['total_count'])

        r = c.get('/api/my-resource/?age__lte=21')
        data = simplejson.loads(r.content)
        self.assertEquals(200, r.status_code)
        self.assertEquals(2, len(data['objects']))
        self.assertEquals(2, data['total_count'])

        # check passing multiple vals to something expecting a single
        r = c.get('/api/my-resource/?age=17&age=21')
        data = simplejson.loads(r.content)
        self.assertEquals(200, r.status_code)
        self.assertEquals(1, len(data['objects']))
        self.assertEquals(1, data['total_count'])

        # Delete the object
        r = c.delete(
            '/api/my-resource/%s/' % obj_data['id'],
            content_type='application/json')
        self.assertEquals(200, r.status_code)

        r = c.get('/api/my-resource/%s/' % obj_data['id'])
        self.assertEquals(404, r.status_code)

    url_conf = 'sprocket.test.test_django_model_resource'

    def setUp(self):
        super(SimpleCase, self).setUp()
        self.org_urls = settings.ROOT_URLCONF
        settings.ROOT_URLCONF = self.url_conf
        _create_table()

    def tearDown(self):
        super(SimpleCase, self).tearDown()
        settings.ROOT_URLCONF = self.org_urls


class FakeModel(Model):
    id = IntegerField(primary_key=True)
    label = CharField()
    created = DjDateTimeField(auto_now_add=True)
    updated = DjDateTimeField(auto_now_add=False)
    published_at = DjDateTimeField()
    email = EmailField()
    age = IntegerField()

    def save(self):
        if not self.pk:
            self.pk = long(time.time() * 1000)
        super(FakeModel, self).save()

    class Meta:
        db_table = 'sprocket_test_fake_model'


class MyModelResource(DjangoModelResource):
    class Meta(ResourceMeta):
        resource_name = 'my-resource'
        model_class = FakeModel
        filtering = {
            'email': ['exact'],
            'age': ['exact', 'range', 'gt', 'gte', 'lt', 'lte', 'in'],
            }

    def on_authenticate(self, request):
        pass


def _create_table():
    sql1 = """DROP TABLE IF EXISTS sprocket_test_fake_model"""
    sql2 = """
    CREATE TABLE sprocket_test_fake_model (
        id INTEGER PRIMARY KEY ASC,
        label,
        published_at DATETIME,
        created DATETIME,
        updated DATETIME,
        email,
        age INTEGER);
    """
    cursor = connections['default'].cursor()
    cursor.execute(sql1, ())
    cursor.execute(sql2, ())
    transaction.commit_unless_managed()
    return cursor.rowcount

my_resource = MyModelResource()

urlpatterns = patterns('',
    (r'^api/', include(my_resource.urls)),
)
